@@ -3,11 +3,11 @@ import random
 
 from functools import wraps
 from math import isinf, isnan, isqrt, sqrt
-from typing import Any, AsyncIterator, Awaitable, Callable, Iterator, Optional, Sequence, SupportsIndex, TypedDict, Union
+from typing import Any, AsyncIterator, Awaitable, Callable, Iterator, Optional, Sequence, SupportsFloat, SupportsIndex, Tuple, TypedDict, Union
 
 import numpy as np
 
-__all__ = ["maximize", "optimize", "optimize_iterator"]
+__all__ = ["maximize", "optimize", "optimize_iterator", "with_input_noise"]
 
 ArrayLike = Union[
     np.ndarray,
@@ -110,6 +110,27 @@ def maximize(f: Callable[[np.ndarray], float], /) -> Callable[[np.ndarray], floa
         return -f(x)
     return wrapper
 
+def with_input_noise(f: Callable[[np.ndarray], float], /, noise: float) -> Callable[[np.ndarray], float]:
+    """Adds noise to the input before calling."""
+    if not callable(f):
+        raise TypeError(f"f must be callable, got {f!r}")
+    elif not isinstance(noise, SupportsFloat):
+        raise TypeError(f"noise must be real, got {noise!r}")
+    noise = float(noise)
+    rng = np.random.default_rng()
+    def rng_iterator(shape: Tuple[int, ...]) -> Iterator[np.ndarray]:
+        while True:
+            random_noise = rng.uniform(-noise, noise, shape)
+            yield random_noise
+            yield random_noise
+    rng_iter: Optional[Iterator[float]] = None
+    def wrapper(x: np.ndarray) -> float:
+        nonlocal rng_iter
+        if rng_iter is None:
+            rng_iter = rng_iterator(x.shape)
+        return f(x + next(rng_iter))
+    return wrapper
+
 def optimize(
     f: Callable[[np.ndarray], float],
     x: ArrayLike,
@@ -128,20 +149,96 @@ def optimize(
     epsilon: float = 1e-7,
 ) -> np.ndarray:
     """
-    Implementation of the SPSA optimization algorithm with adaptive momentum (Adam) and learning rate tuning (line search).
+    Implementation of the SPSA optimization algorithm.
 
-    See `help(spsa)` for more details.
+    Defining Objective Functions
+    -----------------------------
+    Defining your objective function f appropriately can significantly change how well SPSA performs.
+    Here are some tips and tricks on how to define good objective functions.
+
+        Detectable Changes:
+            SPSA requires changes in inputs to lead to changes in outputs.
+            Otherwise SPSA will be unable to determine how inputs should change.
+
+            This isn't possible for all problems. Some may require other algorithms
+            e.g. Bayesian optimization or genetic algorithms. Some may need specialized
+            algorithms for the specific situation.
+
+            If the objective function is noisy then SPSA will increase its change in input
+            until a significant change in output is detected i.e. the result isn't just noise.
+
+            NOTE: SPSA assumes that the result y = f(x) has at least 1e-64 * y noise, so if
+                  there is no change in y, then the change in x automatically increases to try
+                  to search for significant changes. In some cases, this may be sufficient.
+
+        Basin-Hopping Input Noise:
+            Some functions have many local minima, causing SPSA and similar methods to run
+            into bad solutions. This can, to some extent, be countered using `spsa.with_input_noise`.
+
+                def f(x):
+                    return ...
+
+                x = spsa.optimize(spsa.with_input_noise(f, noise=0.5), ...)
+
+            In this way, SPSA will explore neighboring inputs instead of getting stuck in a "basin".
+            If the noise is sufficiently high, and there is a general trend in the direction of the
+            basins towards the best basin, then this will converge to the locally best basin.
+
+            NOTE: Not all functions perform better with input noise, some may perform worse.
+
+        Stochastic vs Deterministic Functions:
+            SPSA does not require deterministic functions.
+
+            If it is more efficient to use a stochastic (random) function e.g. by using a random
+            sample in a dataset each time instead of the whole dataset, then do that instead.
+
+            High-precision accuracy is not possible in this case, but you will get more out of
+            your calculations this way.
+
+            Running a few iterations afterwards with a deterministic variant, if reasonable, can
+            be used to get high-precision accuracy if needed.
+
+        Controllable Stochastic Functions:
+            If the "noise" in the function can be controlled by a second parameter, then the
+            performance of SPSA can be significantly improved by making every pair of calls
+            use the same "noise parameter". This can be implemented as follows:
+
+                from functools import partial
+
+                def rng_iterator():
+                    while True:
+                        random_value = ...  # A sample from a dataset for example.
+                        yield random_value
+                        yield random_value
+
+                def f(x, rng):
+                    random_value = next(rng)
+                    return ...
+
+                x = spsa.optimize(partial(f, rng=rng_iterator()), ...)
+
+        Twice Differentiable:
+            It is not required, but it would be good if the objective function is twice
+            differentiable. SPSA assumes the objective function is relatively smooth in
+            order to converge well. Otherwise it will struggle to move around points
+            where the function is not very smooth.
+
+            For stochastic functions, it is expected that the expected value is smooth.
 
     Parameters
     -----------
         f:
             The function being optimized. Called as `f(array) -> float`.
+
         x:
             The initial point used. This value is edited and returned.
+
         adam:
             True to use Adam, False to not use it.
+
         iterations:
             The number of iterations ran.
+
         lr:
         lr_decay:
         lr_power:
@@ -154,6 +251,7 @@ def optimize(
 
             Furthermore, the learning rate is automatically tuned every iteration to produce
             improved convergence and allow flexible learning rates.
+
         px:
         px_decay:
         px_power:
@@ -168,11 +266,14 @@ def optimize(
 
             Furthermore, the perturbation size is automatically tuned every iteration to produce
             more accurate gradient approximations, reducing chaotic behavior.
+
         momentum:
             The momentum controls how much of the gradient is kept from previous iterations.
+
         beta:
             A secondary momentum, which should be much closer to 1 than the other momentum.
             This is used by the Adam method.
+
         epsilon:
             Used to avoid division by 0 in the Adam method.
 
@@ -213,30 +314,36 @@ def optimize(
             # Increase `px` until the change in f(x) is signficiantly larger than the noise.
             while True:
                 # Update the noise.
-                temp = f(x)
+                y1 = f(x)
+                y2 = f(x)
                 bn += m2 * (1 - bn)
-                y += m2 * (temp - y)
-                noise += m2 * ((temp - f(x)) ** 2 - noise)
+                y += 0.5 * m2 * ((y1 - y) + (y2 - y))
+                noise += m2 * ((y1 - y2) ** 2 - noise)
                 # Compute a change in f(x) in a random direction.
                 dx = rng.choice((-1.0, 1.0), x.shape)
                 dx *= px
                 # Stop if sufficiently accurate.
-                if (f(x + dx) - f(x - dx)) ** 2 > 8 * noise / bn:
+                y1 = f(x + dx)
+                y2 = f(x - dx)
+                if (y1 - y2) ** 2 > 8 * noise / bn or px > 1e-8 + np.linalg.norm(x):
                     break
                 # `dx` is dangerously small, so `px` should be increased.
                 px *= 1.2
             # Attempt to decrease `px` to improve the gradient estimate unless the noise is too much.
             for _ in range(3):
                 # Update the noise.
-                temp = f(x)
+                y1 = f(x)
+                y2 = f(x)
                 bn += m2 * (1 - bn)
-                y += m2 * (temp - y)
-                noise += m2 * ((temp - f(x)) ** 2 - noise)
+                y += 0.5 * m2 * ((y1 - y) + (y2 - y))
+                noise += m2 * ((y1 - y2) ** 2 - noise)
                 # Compute a change in f(x) in a random direction.
                 dx = rng.choice((-1.0, 1.0), x.shape)
                 dx *= px
                 # Stop if too much noise.
-                if (f(x + dx) - f(x - dx)) ** 2 < 8 * noise / bn:
+                y1 = f(x + dx)
+                y2 = f(x - dx)
+                if (y1 - y2) ** 2 < 8 * noise / bn:
                     break
                 # `dx` can be safely decreased, so `px` should be decreased.
                 px /= 1.1
@@ -292,7 +399,9 @@ def optimize(
         dx = rng.choice((-1.0, 1.0), x.shape)
         dx *= px / (1 + px_decay * i) ** px_power
         dx /= np.sqrt(square_gx / b2 + epsilon)
-        df = (f(x_next + dx) - f(x_next - dx)) / 2
+        y1 = f(x_next + dx)
+        y2 = f(x_next - dx)
+        df = (y1 - y2) / 2
         df_dx = dx * (df * sqrt(x.size) / np.linalg.norm(dx) ** 2)
         # Update the gradients.
         b1 += m1 * (1 - b1)
@@ -304,27 +413,29 @@ def optimize(
         dx = gx / (b1 * (1 + lr_decay * i) ** lr_power)
         if adam:
             dx /= np.sqrt(square_gx / b2 + epsilon)
+        # Sample points.
+        y3 = f(x)
+        y4 = f(x - lr / 3 * dx)
+        y5 = f(x - lr * 3 * dx)
+        y6 = f(x)
         # Estimate the noise in f.
-        y1 = f(x)
-        y2 = f(x - lr / 3 * dx)
-        y3 = f(x - lr * 3 * dx)
         bn += m2 * (1 - bn)
-        y += m2 * (y1 - y)
-        noise += m2 * ((y1 - f(x)) ** 2 - noise)
+        y += m2 * (y3 - y)
+        noise += m2 * ((y3 - y6) ** 2 + 1e-64 * (abs(y3) + abs(y6)) - noise)
         # Update `px` depending on the noise and gradient.
         # `dx` is dangerously small, so `px` should be increased.
-        if df ** 2 < 2 * noise / bn:
+        if (y1 - y2) ** 2 < 8 * noise / bn and px < 1e-8 + np.linalg.norm(x):
             px *= 1.2
         # `dx` can be safely decreased, so `px` should be decreased.
         elif px > 1e-8 * (1 + 0.25 * np.linalg.norm(x)):
             px /= 1.1
         # Perform line search.
         # Adjust the learning rate towards learning rates which give good results.
-        if y1 - 0.25 * sqrt(noise / bn) < min(y2, y3):
+        if y3 - 0.25 * sqrt(noise / bn) < min(y4, y5):
             lr /= 1.3
-        if y2 - 0.25 * sqrt(noise / bn) < min(y1, y3):
+        if y4 - 0.25 * sqrt(noise / bn) < min(y3, y5):
             lr *= 1.3 / 1.4
-        if y3 - 0.25 * sqrt(noise / bn) < min(y1, y2):
+        if y5 - 0.25 * sqrt(noise / bn) < min(y3, y4):
             lr *= 1.4
         # Set a minimum learning rate.
         lr = max(lr, epsilon / (1 + 0.01 * i) ** 0.5 * (1 + 0.25 * np.linalg.norm(x)))
@@ -386,17 +497,22 @@ def optimize_iterator(
         optimizer_variables:
             A dictionary containing optimizer variables.
 
+            NOTE: Most variables come with a corresponding beta variable which should be used.
+
             NOTE: x, gradient, slow_gradient, and square_gradient are mutable numpy arrays.
                   Modifying them may mess up the optimizer.
 
             x_min:
             y_min:
                 The best seen estimated minimum of f.
+
             x:
             y:
                 The current estimated minimum of f.
+
             lr:
                 The current learning rate (not including decay).
+
             beta_x
             beta_noise:
             beta1:
@@ -412,15 +528,19 @@ def optimize_iterator(
 
                 On their own, the estimates are closer to 0 than they should be
                 and need to be divided by their respective betas for correction.
+
             noise:
                 An estimate for how much noise is in f(x).
                 Used for SPSA.
+
             gradient:
                 The estimated gradient of f at x.
+
             slow_gradient:
                 The slower estimate of the gradient of f at x.
                 Biased more towards previous iterations.
                 Used for the square_gradient.
+
             square_gradient:
                 An estimate for the component-wise square of the gradient of f at x.
                 Used for the Adam method and perturbation size rescaling.
@@ -457,30 +577,36 @@ def optimize_iterator(
             # Increase `px` until the change in f(x) is signficiantly larger than the noise.
             while True:
                 # Update the noise.
-                temp = f(x)
+                y1 = f(x)
+                y2 = f(x)
                 bn += m2 * (1 - bn)
-                y += m2 * (temp - y)
-                noise += m2 * ((temp - f(x)) ** 2 - noise)
+                y += 0.5 * m2 * ((y1 - y) + (y2 - y))
+                noise += m2 * ((y1 - y2) ** 2 + 1e-64 * (abs(y1) + abs(y2)) - noise)
                 # Compute a change in f(x) in a random direction.
                 dx = rng.choice((-1.0, 1.0), x.shape)
                 dx *= px
                 # Stop if sufficiently accurate.
-                if (f(x + dx) - f(x - dx)) ** 2 > 8 * noise / bn:
+                y1 = f(x + dx)
+                y2 = f(x - dx)
+                if (y1 - y2) ** 2 > 8 * noise / bn or px > 1e-8 + np.linalg.norm(x):
                     break
                 # `dx` is dangerously small, so `px` should be increased.
                 px *= 1.2
             # Attempt to decrease `px` to improve the gradient estimate unless the noise is too much.
             for _ in range(3):
                 # Update the noise.
-                temp = f(x)
+                y1 = f(x)
+                y2 = f(x)
                 bn += m2 * (1 - bn)
-                y += m2 * (temp - y)
-                noise += m2 * ((temp - f(x)) ** 2 - noise)
+                y += 0.5 * m2 * ((y1 - y) + (y2 - y))
+                noise += m2 * ((y1 - y2) ** 2 + 1e-64 * (abs(y1) + abs(y2)) - noise)
                 # Compute a change in f(x) in a random direction.
                 dx = rng.choice((-1.0, 1.0), x.shape)
                 dx *= px
                 # Stop if too much noise.
-                if (f(x + dx) - f(x - dx)) ** 2 < 8 * noise / bn:
+                y1 = f(x + dx)
+                y2 = f(x - dx)
+                if (y1 - y2) ** 2 < 8 * noise / bn:
                     break
                 # `dx` can be safely decreased, so `px` should be decreased.
                 px /= 1.1
@@ -554,40 +680,43 @@ def optimize_iterator(
         dx = rng.choice((-1.0, 1.0), x.shape)
         dx *= px / (1 + px_decay * i) ** px_power
         dx /= np.sqrt(square_gx / b2 + epsilon)
-        df = (f(x_next + dx) - f(x_next - dx)) / 2
+        y1 = f(x_next + dx)
+        y2 = f(x_next - dx)
+        df = (y1 - y2) / 2
         df_dx = dx * (df * sqrt(x.size) / np.linalg.norm(dx) ** 2)
         # Update the gradients.
         b1 += m1 * (1 - b1)
+        b2 += m2 * (1 - b2)
         gx += m1 * (df_dx - gx)
-        if adam:
-            b2 += m2 * (1 - b2)
-            slow_gx += m2 * (df_dx - slow_gx)
-            square_gx += m2 * ((slow_gx / b2) ** 2 - square_gx)
+        slow_gx += m2 * (df_dx - slow_gx)
+        square_gx += m2 * ((slow_gx / b2) ** 2 - square_gx)
         # Compute the step size.
         dx = gx / (b1 * (1 + lr_decay * i) ** lr_power)
         if adam:
             dx /= np.sqrt(square_gx / b2 + epsilon)
+        # Sample points.
+        y3 = f(x)
+        y4 = f(x - lr / 3 * dx)
+        y5 = f(x - lr * 3 * dx)
+        y6 = f(x)
         # Estimate the noise in f.
-        y1 = f(x)
-        y2 = f(x - lr / 3 * dx)
-        y3 = f(x - lr * 3 * dx)
         bn += m2 * (1 - bn)
-        y += m2 * (y1 - y)
-        noise += m2 * ((y1 - f(x)) ** 2 - noise)
+        y += m2 * (y3 - y)
+        noise += m2 * ((y3 - y6) ** 2 + 1e-64 * (abs(y3) + abs(y6)) - noise)
         # Update `px` depending on the noise and gradient.
         # `dx` is dangerously small, so `px` should be increased.
-        if df ** 2 < 2 * noise / bn:
+        if (y1 - y2) ** 2 < 8 * noise / bn and px < 1e-8 + np.linalg.norm(x):
             px *= 1.2
         # `dx` can be safely decreased, so `px` should be decreased.
         elif px > 1e-8 * (1 + 0.25 * np.linalg.norm(x)):
             px /= 1.1
         # Perform line search.
         # Adjust the learning rate towards learning rates which give good results.
-        if y1 - 0.25 * sqrt(noise / bn) < min(y2, y3):
+        if y3 - 0.25 * sqrt(noise / bn) < min(y4, y5):
             lr /= 1.3
-        if y2 - 0.25 * sqrt(noise / bn) < min(y1, y3):
+        if y4 - 0.25 * sqrt(noise / bn) < min(y3, y5):
             lr *= 1.3 / 1.4
-        if y3 - 0.25 * sqrt(noise / bn) < min(y1, y2):
+        if y5 - 0.25 * sqrt(noise / bn) < min(y3, y4):
             lr *= 1.4
         # Set a minimum learning rate.
         lr = max(lr, epsilon / (1 + 0.01 * i) ** 0.5 * (1 + 0.25 * np.linalg.norm(x)))
