@@ -41,13 +41,13 @@ import operator
 import random
 
 from math import isinf, isnan, isqrt, sqrt
-from typing import AsyncIterator, Awaitable, Callable, Optional, Sequence, SupportsIndex
+from typing import AsyncIterator, Awaitable, Callable, Optional, Sequence, SupportsFloat, SupportsIndex, Tuple
 
 import numpy as np
 
 from ._spsa import ArrayLike, OptimizerVariables, _type_check
 
-__all__ = ["optimize", "optimize_iterator"]
+__all__ = ["maximize", "optimize", "optimize_iterator", "with_input_noise"]
 
 async def maximize(f: Callable[[np.ndarray], Awaitable[float]], /) -> Callable[[np.ndarray], Awaitable[float]]:
     """
@@ -63,6 +63,27 @@ async def maximize(f: Callable[[np.ndarray], Awaitable[float]], /) -> Callable[[
         return -(await f(x))
     return wrapper
 
+def with_input_noise(f: Callable[[np.ndarray], Awaitable[float]], /, noise: float) -> Callable[[np.ndarray], Awaitable[float]]:
+    """Adds noise to the input before calling."""
+    if not callable(f):
+        raise TypeError(f"f must be callable, got {f!r}")
+    elif not isinstance(noise, SupportsFloat):
+        raise TypeError(f"noise must be real, got {noise!r}")
+    noise = float(noise)
+    rng = np.random.default_rng()
+    def rng_iterator(shape: Tuple[int, ...]) -> Iterator[np.ndarray]:
+        while True:
+            random_noise = rng.uniform(-noise, noise, shape)
+            yield random_noise
+            yield random_noise
+    rng_iter: Optional[Iterator[float]] = None
+    async def wrapper(x: np.ndarray) -> float:
+        nonlocal rng_iter
+        if rng_iter is None:
+            rng_iter = rng_iterator(x.shape)
+        return await f(x + next(rng_iter))
+    return wrapper
+
 async def optimize(
     f: Callable[[np.ndarray], Awaitable[float]],
     x: ArrayLike,
@@ -122,13 +143,13 @@ async def optimize(
                 y1, y2 = await asyncio.gather(f(x), f(x))
                 bn += m2 * (1 - bn)
                 y += 0.5 * m2 * ((y1 - y) + (y2 - y))
-                noise += m2 * ((y1 - y2) ** 2 - noise)
+                noise += m2 * ((y1 - y2) ** 2 + 1e-64 * (abs(y1) + abs(y2)) - noise)
                 # Compute a change in f(x) in a random direction.
                 dx = rng.choice((-1.0, 1.0), x.shape)
                 dx *= px
                 # Stop if sufficiently accurate.
                 y1, y2 = await asyncio.gather(f(x + dx), f(x - dx))
-                if (y1 - y2) ** 2 > 8 * noise / bn:
+                if (y1 - y2) ** 2 > 8 * noise / bn or px > 1e-8 + np.linalg.norm(x):
                     break
                 # `dx` is dangerously small, so `px` should be increased.
                 px *= 1.2
@@ -139,7 +160,7 @@ async def optimize(
                 y1, y2 = await asyncio.gather(f(x), f(x))
                 bn += m2 * (1 - bn)
                 y += 0.5 * m2 * ((y1 - y) + (y2 - y))
-                noise += m2 * ((y1 - y2) ** 2 - noise)
+                noise += m2 * ((y1 - y2) ** 2 + 1e-64 * (abs(y1) + abs(y2)) - noise)
                 # Compute a change in f(x) in a random direction.
                 dx = rng.choice((-1.0, 1.0), x.shape)
                 dx *= px
@@ -209,7 +230,6 @@ async def optimize(
         dx = rng.choice((-1.0, 1.0), x.shape)
         dx *= px / (1 + px_decay * i) ** px_power
         dx /= np.sqrt(square_gx / b2 + epsilon)
-        df = (f(x_next + dx) - f(x_next - dx)) / 2
         y1, y2 = await asyncio.gather(f(x_next + dx), f(x_next - dx))
         df = (y1 - y2) / 2
         df_dx = dx * (df * sqrt(x.size) / np.linalg.norm(dx) ** 2)
@@ -223,26 +243,26 @@ async def optimize(
         dx = gx / (b1 * (1 + lr_decay * i) ** lr_power)
         if adam:
             dx /= np.sqrt(square_gx / b2 + epsilon)
-        # Sample points in parallel.
-        y1, y2, y3, y0 = await asyncio.gather(f(x), f(x - lr / 3 * dx), f(x - lr * 3 * dx), f(x))
+        # Sample points concurrently.
+        y3, y4, y5, y6 = await asyncio.gather(f(x), f(x - lr / 3 * dx), f(x - lr * 3 * dx), f(x))
         # Estimate the noise in f.
         bn += m2 * (1 - bn)
-        y += 0.5 * m2 * ((y0 - y) + (y1 - y))
-        noise += m2 * ((y0 - y1) ** 2 - noise)
+        y += m2 * (y3 - y)
+        noise += m2 * ((y3 - y6) ** 2 + 1e-64 * (abs(y3) + abs(y6)) - noise)
         # Update `px` depending on the noise and gradient.
         # `dx` is dangerously small, so `px` should be increased.
-        if df ** 2 < 2 * noise / bn:
+        if (y1 - y2) ** 2 < 8 * noise / bn and px < 1e-8 + np.linalg.norm(x):
             px *= 1.2
         # `dx` can be safely decreased, so `px` should be decreased.
         elif px > 1e-8 * (1 + 0.25 * np.linalg.norm(x)):
             px /= 1.1
         # Perform line search.
         # Adjust the learning rate towards learning rates which give good results.
-        if y1 - 0.25 * sqrt(noise / bn) < min(y2, y3):
+        if y3 - 0.25 * sqrt(noise / bn) < min(y4, y5):
             lr /= 1.3
-        if y2 - 0.25 * sqrt(noise / bn) < min(y1, y3):
+        if y4 - 0.25 * sqrt(noise / bn) < min(y3, y5):
             lr *= 1.3 / 1.4
-        if y3 - 0.25 * sqrt(noise / bn) < min(y1, y2):
+        if y5 - 0.25 * sqrt(noise / bn) < min(y3, y4):
             lr *= 1.4
         # Set a minimum learning rate.
         lr = max(lr, epsilon / (1 + 0.01 * i) ** 0.5 * (1 + 0.25 * np.linalg.norm(x)))
@@ -335,13 +355,13 @@ async def optimize_iterator(
                 y1, y2 = await asyncio.gather(f(x), f(x))
                 bn += m2 * (1 - bn)
                 y += 0.5 * m2 * ((y1 - y) + (y2 - y))
-                noise += m2 * ((y1 - y2) ** 2 - noise)
+                noise += m2 * ((y1 - y2) ** 2 + 1e-64 * (abs(y1) + abs(y2)) - noise)
                 # Compute a change in f(x) in a random direction.
                 dx = rng.choice((-1.0, 1.0), x.shape)
                 dx *= px
                 # Stop if sufficiently accurate.
                 y1, y2 = await asyncio.gather(f(x + dx), f(x - dx))
-                if (y1 - y2) ** 2 > 8 * noise / bn:
+                if (y1 - y2) ** 2 > 8 * noise / bn or px > 1e-8 + np.linalg.norm(x):
                     break
                 # `dx` is dangerously small, so `px` should be increased.
                 px *= 1.2
@@ -352,7 +372,7 @@ async def optimize_iterator(
                 y1, y2 = await asyncio.gather(f(x), f(x))
                 bn += m2 * (1 - bn)
                 y += 0.5 * m2 * ((y1 - y) + (y2 - y))
-                noise += m2 * ((y1 - y2) ** 2 - noise)
+                noise += m2 * ((y1 - y2) ** 2 + 1e-64 * (abs(y1) + abs(y2)) - noise)
                 # Compute a change in f(x) in a random direction.
                 dx = rng.choice((-1.0, 1.0), x.shape)
                 dx *= px
@@ -453,26 +473,26 @@ async def optimize_iterator(
         dx = gx / (b1 * (1 + lr_decay * i) ** lr_power)
         if adam:
             dx /= np.sqrt(square_gx / b2 + epsilon)
-        # Sample points in parallel.
-        y1, y2, y3, y0 = await asyncio.gather(f(x), f(x - lr / 3 * dx), f(x - lr * 3 * dx), f(x))
+        # Sample points concurrently.
+        y3, y4, y5, y6 = await asyncio.gather(f(x), f(x - lr / 3 * dx), f(x - lr * 3 * dx), f(x))
         # Estimate the noise in f.
         bn += m2 * (1 - bn)
-        y += 0.5 * m2 * ((y0 - y) + (y1 - y))
-        noise += m2 * ((y0 - y1) ** 2 - noise)
+        y += m2 * (y3 - y)
+        noise += m2 * ((y3 - y6) ** 2 + 1e-64 * (abs(y3) + abs(y6)) - noise)
         # Update `px` depending on the noise and gradient.
         # `dx` is dangerously small, so `px` should be increased.
-        if df ** 2 < 2 * noise / bn:
+        if (y1 - y2) ** 2 < 8 * noise / bn and px < 1e-8 + np.linalg.norm(x):
             px *= 1.2
         # `dx` can be safely decreased, so `px` should be decreased.
         elif px > 1e-8 * (1 + 0.25 * np.linalg.norm(x)):
             px /= 1.1
         # Perform line search.
         # Adjust the learning rate towards learning rates which give good results.
-        if y1 - 0.25 * sqrt(noise / bn) < min(y2, y3):
+        if y3 - 0.25 * sqrt(noise / bn) < min(y4, y5):
             lr /= 1.3
-        if y2 - 0.25 * sqrt(noise / bn) < min(y1, y3):
+        if y4 - 0.25 * sqrt(noise / bn) < min(y3, y5):
             lr *= 1.3 / 1.4
-        if y3 - 0.25 * sqrt(noise / bn) < min(y1, y2):
+        if y5 - 0.25 * sqrt(noise / bn) < min(y3, y4):
             lr *= 1.4
         # Set a minimum learning rate.
         lr = max(lr, epsilon / (1 + 0.01 * i) ** 0.5 * (1 + 0.25 * np.linalg.norm(x)))
